/*
** Copyright (c) 1997, 1998, 1999
**	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
**
*/
/*****************************************************************************
** sdla_dev.h	SDLA ISA/PCI Header.
**
** Author:	Alex Feldman	<al.feldman@sangoma.com>
**
*****************************************************************************
*/

#ifndef __SDLA_DEV_H
# define __SDLA_DEV_H

#if defined(__FreeBSD__)
# if (__FreeBSD_version < 502000)
#  include <pci/pcireg.h>
#  include <pci/pcivar.h>
# else
#  include <dev/pci/pcireg.h>
#  include <dev/pci/pcivar.h>
# endif
#elif defined(__NetBSD__) || defined(__OpenBSD__)
# include <dev/pci/pcireg.h>
# include <dev/pci/pcivar.h>
# include <dev/isa/isavar.h>
#else
# error "This operation system not supported yet!"
#endif


/*
**************************************************************************
**				DEFINES/MACROS				**
**************************************************************************
*/
#ifdef __SDLA_DEV_C
# define EXTERN
#else
# define EXTERN extern
#endif

/* the same value as in sdladrv.h */
#define SDLA_MAX_PORTS	8

/* Offset from I/O port for S508 ID */
#define SDLA_S508_OFFSET	0x3

#define SDLA_DEV_NAME "sdla"

/* SDLA ISA device name */
#define SDLA_ISA_DEV_NAME "sdlai"

/* SDLA PCI device name */
#define SDLA_PCI_DEV_NAME "sdlap"

/* I/O ports */
#define _INB(port)       (inb(port))
#define _OUTB(port,byte) (outb(port,byte))

/* read a value off base + _offset */
#define s508_read_port(_addr, _offset) _INB(_addr + _offset)

#define sdladev_slot(sdladev)	sdladev->u.pci.slot_no
#define sdladev_bus(sdladev)	sdladev->u.pci.bus_no
#define sdladev_cpu(sdladev)	sdladev->u.pci.cpu_no
#define sdladev_irq(sdladev)	sdladev->irq
#define sdladev_ioport(sdladev)	sdladev->u.isa.ioport
#if defined(__FreeBSD__)
# define sdladev_maddr(sdladev)	sdladev->u.isa.maddr
#elif defined(__NetBSD__) || defined(__OpenBSD__)
# define sdladev_maddr(sdladev)	sdladev->u.isa.memh
#endif


/*
**************************************************************************
**				TYPEDEF					**
**************************************************************************
*/
typedef struct sdla_intr_arg_ {
	int	ready;
	void   	(*isr)(void *arg); 	/* Interrupt handler */
	void	*arg;			/* Interrupt args */
	void	*sd;
	int	line_no;
} sdla_intr_arg_t;

/****** SDLA ISA/PCI Info structure *****************************************/
#if defined(__NetBSD__) || defined(__OpenBSD__) || (defined(__FreeBSD__) && (__FreeBSD_version > 400000))
struct sdla_softc
{
# if defined(__FreeBSD__)
	struct device* 	dev;		/* Pointer to device structure */
	int 		irq_rid;
	struct resource* irq_res;
#  if (__FreeBSD_version > 400000)
	int		mem_rid;
	int		mem_used;
	struct resource* mem_res;
	int		port_rid;
	int 		port_used;
	struct resource* port_res;
#  endif
# elif defined(__NetBSD__) || defined(__OpenBSD__)
	struct device		dev;
	struct isa_attach_args	ia;
	struct pci_attach_args 	pa;
	/* ISA interrupt handler */
	void*			sc_ih;
# endif
};
#endif

typedef struct {
	u_short	ioport;
#if defined(__FreeBSD__)
	caddr_t	maddr;
	int 	isa_unit;
#elif defined(__NetBSD__) || defined(__OpenBSD__)
	bus_space_handle_t 	memh;	/* Memory address handle */
	bus_space_handle_t 	ioh;	/* I/O port handle */
	bus_space_tag_t 	memt;	/* Memory address tag 	 */
	bus_space_tag_t 	iot;	/* I/O port tag */
#endif
} sdla_isa_t;

typedef struct {
	u_int16_t	cpu_no;		/* CPU number */
	u_int16_t	line_no;	/* Line number AAA */
#if defined(__FreeBSD__)
# if (__FreeBSD_version > 400000)
	u_int8_t	slot_no;	/* Slot number */
	u_int8_t	bus_no;		/* Bus number */
# else
	pcici_t		pci_dev;	/* PCI cfg struct */
#  define slot_no	pci_dev->slot
#  define bus_no	pci_dev->bus
# endif
#elif defined(__NetBSD__) || defined(__OpenBSD__)
	u_int8_t	slot_no;	/* Slot number */
	u_int8_t	bus_no;		/* Bus number */
#endif
} sdla_pci_t;

typedef struct
{
	int  		type;		/* SDLA_S508/SDLA_S514/SDLA_ADSL */
	int  		adapter_type;	/* FE adapter type */
	int  		adapter_subtype;/* FE adapter subtype (NORMAL/SHARK) */
	u_int16_t	vendor_id;	/* Vendor id */
	u_int16_t	subvendor_id;	/* SubVendor id */
	u_int16_t	subsystem_id;	/* SubSystem id */
	u_int16_t	device_id;	/* Device id */
#if defined(__FreeBSD__) && (__FreeBSD_version >= 450000)
	device_t	dev;
#else
	sdla_intr_arg_t	intr_arg[SDLA_MAX_PORTS];
#endif
	void*		irqh;
	int		irq;
	union {
		sdla_isa_t	isa;
		sdla_pci_t	pci;
	} u;
#if defined(__NetBSD__) || defined(__OpenBSD__) || (__FreeBSD_version > 400000)
	struct sdla_softc*	sc;	/* Pointer to sdla_softc structure */
#endif
} sdladev_t;

#if defined(__NetBSD__) || defined(__OpenBSD__)
EXTERN int	  sdla_probe(int);
EXTERN sdladev_t* sdla_attach(struct sdla_softc*, int);
#endif

#undef EXTERN

#endif /* __SDLA_DEV_H */

/****** End *****************************************************************/
