/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: wanmain.c,v 1.37 2006/04/18 19:50:05 sangoma Exp $
 */

/*****************************************************************************
 * wanmain.c	WAN Multiprotocol Router Module. Main code.
 *
 *		This module is completely hardware-independent and provides
 *		the following common services for the WAN Link Drivers:
 *		 o WAN device managenment (registering, unregistering)
 *		 o Network interface management
 *		 o Physical connection management (dial-up, incomming calls)
 *		 o Logical connection management (switched virtual circuits)
 *
 * Author:	Alex Feldman	<al.feldmang@sangoma.com>
 *
 * ============================================================================
 * Dec 12, 1999	Alex Feldman	Updated for FreeBSD 3.2 
 * Dec 27, 1996	Gene Kozin	Initial version (based on Sangoma's WANPIPE)
 ******************************************************************************
 */

/*
 ****************************************************************************** 
 *				INCLUDE FILES 				      *
 ******************************************************************************
 */ 
#include <net/wanpipe_version.h>
#include <net/wanpipe_includes.h>
#include <net/wanpipe_defines.h>
#include <net/wanpipe_debug.h>
#include <net/wanpipe_common.h>
#include <net/wanpipe_iface.h>
#include <net/if_wanpipe_common.h>
#include <net/wanproc.h>
#include <net/wanpipe.h>
#include <net/wanpipe_lip_kernel.h>

/****************************************************************************** 
 *				DEFINES and MACROS			      *
 ******************************************************************************
 */ 

#ifndef	min
#define min(a,b) (((a)<(b))?(a):(b))
#endif
#ifndef	max
#define max(a,b) (((a)>(b))?(a):(b))
#endif

/*
 ****************************************************************************** 
 *			TYPEDEF and STRUCTURES				      *
 ******************************************************************************
 */ 
#if defined(__FreeBSD__)
d_open_t wanrouter_open;
d_close_t wanrouter_close;
d_ioctl_t wanrouter_ioctl;

# if (__FreeBSD_version >= 600000)
static struct cdevsw wanrouter_devsw = 
{
	.d_version	= D_VERSION,
	.d_flags	= D_NEEDGIANT,
	.d_name		= "wanrouter",
	.d_open		= wanrouter_open,
	.d_close	= wanrouter_close,
	.d_ioctl	= wanrouter_ioctl,
};
# elif (__FreeBSD_version >= 503000)
static struct cdevsw wanrouter_devsw = 
{
	.d_version	= D_VERSION,
	.d_maj		= WAN_CDEV_MAJOR, 
	.d_flags	= D_NEEDGIANT,
	.d_name		= "wanrouter",
	.d_open		= wanrouter_open,
	.d_close	= wanrouter_close,
	.d_ioctl	= wanrouter_ioctl,
};
# elif (__FreeBSD_version >= 501000)
static struct cdevsw wanrouter_devsw = 
{
	WAN_CDEV_MAJOR,
	D_NOGIANT,	/* FIXME: ???? */
	"wanrouter",
	wanrouter_open,
# if (__FreeBSD_version >= 502000)
	(d_fdopen_t*)NULL,
# endif
	wanrouter_close,
	(d_read_t*)NULL,
	(d_write_t*)NULL,
	wanrouter_ioctl,
	(d_poll_t*)NULL,
	(d_mmap_t*)NULL,
	(d_strategy_t*)NULL,
	NULL,
};
# elif (__FreeBSD_version > 400000)
static struct cdevsw wanrouter_devsw = 
{
	wanrouter_open,
	wanrouter_close,
	(d_read_t*)NULL,
	(d_write_t*)NULL,
	wanrouter_ioctl,
	(d_poll_t*)NULL,
	(d_mmap_t*)NULL,
	(d_strategy_t*)NULL,
	"wanrouter",
	WAN_CDEV_MAJOR,
	NULL,
	nopsize,
	0,
	0
};
# else
static struct cdevsw wanrouter_devsw = 
{
	wanrouter_open,
	wanrouter_close,
	noread,
	nowrite,
	wanrouter_ioctl,
	nostop,
	nullreset,
	nodevtotty,
	NULL,
	nommap,
	NULL,
	"wanrouter",
	NULL,
	WAN_CDEV_MAJOR
};
# endif
#else
int wanrouter_open __P((wan_cdev_t dev, int oflags, int devtype, struct proc* p));
int wanrouter_close __P((wan_cdev_t dev, int oflags, int devtype, struct proc* p));
int wanrouter_ioctl __P((wan_cdev_t dev, u_long cmd, caddr_t data, int fflag, wan_proc_ptr_t procp));

struct cdevsw wanrouter_devsw = {
	wanrouter_open,
	wanrouter_close,
	NULL,
	NULL,
	wanrouter_ioctl,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};

#endif

/* 
 *	WAN device IOCTL handlers 
 */

static int wan_device_setup(wan_device_t *wandev, wandev_conf_t *u_conf);
static int wan_device_stat(wan_device_t *wandev, wandev_stat_t *u_stat);
static int wan_device_shutdown(wan_device_t *wandev, wandev_conf_t* u_conf);
static int wan_device_debugging(wan_device_t *wandev, char* u_ifname);
static int wan_device_new_if(wan_device_t *wandev, wanif_conf_t *u_conf);
static int wan_device_del_if(wan_device_t *wandev, char *u_name);
static int wan_device_procfs(wan_device_t* wandev, wan_procfs_t* procfs);

int register_wanpipe_lip_protocol (wplip_reg_t *lip_reg);
void unregister_wanpipe_lip_protocol (void);
static int wan_device_unreg_lip(netdevice_t *dev);
static int wan_device_new_if_lip(wan_device_t *wandev, wanif_conf_t *u_conf);
static int wan_device_del_if_lip(wan_device_t *wandev, netdevice_t *dev);

/* 
 *	Miscellaneous 
 */

static wan_device_t *find_device (char *name); 
static void wanpipe_attach(struct ifnet *ifp, wan_device_t* wandev);
static void wanpipe_detach(struct ifnet *ifp, wan_device_t* wandev);
static int delete_interface (wan_device_t *wandev, netdevice_t *dev);


/*
 *	Global Data
 */
struct wan_devlist_	wan_devlist = 
	WAN_LIST_HEAD_INITIALIZER(wan_devlist); 
/*wan_device_t* router_devlist 	= NULL;*/	/* list of registered devices */
static int devcnt 		= 0;
static char wan_fullname[]		= "WANPIPE(tm) Interface Support Module";
static unsigned char wan_version[100];
struct wplip_reg		   wplip_protocol;

/********************* Kernel Loadable Module Entry Points *******************/
int wanrouter_init(void *);
int wanrouter_exit(void *);
#if defined(__FreeBSD__)
# if (__FreeBSD_version > 400000)
static wan_cdev_t	wanpipe_dev;
#endif
WAN_MODULE_DEFINE(
		wanrouter, "wanrouter",
		"Alex Feldman <al.feldman@sangoma.com>",
		"WANPIPE(tm) Interface Support Module",
		"GPL",
		wanrouter_init, wanrouter_exit, 
		&wanrouter_devsw);
MODULE_VERSION(wanrouter,WANROUTER_MAJOR_VER);
MODULE_DEPEND(wanrouter,sdladrv,1,SDLADRV_MAJOR_VER,SDLADRV_MAJOR_VER);
#endif

int wanrouter_init(void *devsw)
{
	WAN_LIST_INIT(&wan_devlist);
	if (WANPIPE_VERSION_BETA){
		DEBUG_EVENT("%s Beta%s-%s %s %s\n",
			wan_fullname, WANPIPE_SUB_VERSION, WANPIPE_VERSION,
			WANPIPE_COPYRIGHT_DATES,WANPIPE_COMPANY);
		sprintf(wan_version,"Beta%s-%s",
				WANPIPE_SUB_VERSION, WANPIPE_VERSION);
	}else{
		DEBUG_EVENT("%s Stable %s-%s %s %s\n",
			wan_fullname, WANPIPE_VERSION, WANPIPE_SUB_VERSION,
			WANPIPE_COPYRIGHT_DATES,WANPIPE_COMPANY);
		sprintf(wan_version,"Stable %s-%s",
				WANPIPE_VERSION, WANPIPE_SUB_VERSION);
	}
	        
	DEBUG_MOD("Loading WANPIPE module...\n");
#if defined(__FreeBSD__) && (__FreeBSD_version > 400000)
	if (devsw){
		wanpipe_dev = make_dev((struct cdevsw*)devsw,
					0,UID_ROOT,GID_WHEEL,
					0600,"wanrouter");
	}
#endif
	UNREG_PROTOCOL_FUNC(wplip_protocol);
	return 0;
}

int wanrouter_exit(void *devsw)
{
#if defined(__FreeBSD__) && (__FreeBSD_version > 400000)
	destroy_dev(wanpipe_dev);
#endif
	return 0;
}

/*
 * 	Kernel APIs
 */

/* 
 * ============================================================================
 * Register WAN device.
 * o verify device credentials
 * o create an entry for the device in the /proc/net/router directory(LINUX)
 * o initialize internally maintained fields of the wan_device structure
 * o link device data space to a singly-linked list
 * o if it's the first device, then start kernel 'thread'
 * o increment module use count
 *
 * Return:
 * 0	Ok
 * < 0	error.
 *
 * Context:	process
 */
int register_wan_device(wan_device_t *wandev)
{
	int namelen;

	WAN_ASSERT(wandev == NULL);
	WAN_ASSERT(wandev->magic != ROUTER_MAGIC);
	WAN_ASSERT(wandev->name == NULL);
 
	namelen = strlen(wandev->name);
	if (!namelen || (namelen > WAN_DRVNAME_SZ)){
		return -EINVAL;
	}
	
	if (find_device(wandev->name) != NULL){
		return -EEXIST;
	}

	DEBUG_INIT("%s: registering WAN device...\n", 
					wandev->name);

	/*
	 * Initialize fields of the wan_device structure 
	 * maintained by the router and update local data.
     	 */
 
	wandev->ndev    = 0;
	WAN_LIST_INIT(&wandev->dev_head);
	WAN_LIST_INSERT_HEAD(&wan_devlist, wandev, next);
/*
	wandev->next    = router_devlist;
	router_devlist  = wandev;*/
	++devcnt;
	return 0;
}

/* 
 * ============================================================================
 *	Unregister WAN device.
 *	o shut down device
 *	o unlink device data space from the linked list
 *	o delete device entry in the /proc/net/router directory
 *	o decrement module use count
 *
 *	Return:		0	Ok
 *			<0	error.
 *	Context:	process
 */

int unregister_wan_device(char *name)
{
	wan_device_t *wandev = NULL;

	WAN_ASSERT(name == NULL);	

	WAN_LIST_FOREACH(wandev, &wan_devlist, next){
		if (!strcmp(wandev->name, name)){
			break;
		}
	}
/*
	for (wandev = router_devlist, prev = NULL;
		wandev && strcmp(wandev->name, name);
		prev = wandev, wandev = wandev->next);
*/
	if (wandev == NULL){
		return -ENODEV;
	}

	DEBUG_INIT("%s: unregistering WAN device\n", 
					wandev->name);

	if (wandev->state != WAN_UNCONFIGURED){ 
		wan_device_shutdown(wandev, NULL);
	}
	WAN_LIST_REMOVE(wandev, next);
/*
    	if (prev){
	        prev->next = wandev->next;
	}else{
		router_devlist = wandev->next;
	}
*/
	--devcnt;
	return 0;
}

int wanrouter_open(wan_cdev_t dev, int flag, int otyp, wan_proc_ptr_t procp)
{
	return (0);
}
int wanrouter_close(wan_cdev_t dev, int flag, int otyp, wan_proc_ptr_t procp)
{   
	return (0);
}
/*
 * ============================================================================
 * 	WAN device IOCTL.
 *	o find WAN device associated with this node
 *	o execute requested action or pass command to the device driver
 */
int wanrouter_ioctl(wan_cdev_t dev, u_long cmd, caddr_t arg, int mode, wan_proc_ptr_t procp)
{
	int err = 0;
	wan_device_t *wandev = NULL;
	wan_conf_t *cfg = NULL;

	WAN_ASSERT(((cmd >> 8) & ROUTER_IOCTL) != ROUTER_IOCTL);
	WAN_ASSERT(arg == NULL);
    
	/* The arg is points already to kernel space */
	cfg = (wan_conf_t*)arg;

	wandev = find_device(cfg->devname);
	if (!((wandev != NULL && wandev->magic == ROUTER_MAGIC) || 
	     (wandev == NULL && cmd == ROUTER_PROCFS))){
        	return -EINVAL;
	}
		
	switch(cmd){ 
    	case ROUTER_SETUP:
		DEBUG_CFG("%s: ROUTER_SETUP \n", cfg->devname);
		err = wan_device_setup(wandev, cfg->arg);
		break;

	case ROUTER_DOWN:
		DEBUG_CFG("%s: ROUTER_DOWN\n", cfg->devname);
		err = wan_device_shutdown(wandev, cfg->arg);
		break;

	case ROUTER_DEBUGGING:
		DEBUG_CFG("%s: ROUTER_DEBUGGING\n", cfg->devname);
		err = wan_device_debugging(wandev, cfg->arg);
		break;

	case ROUTER_STAT:
		DEBUG_CFG("%s: ROUTER_STAT\n", cfg->devname); 
		err = wan_device_stat(wandev, cfg->arg);
		break;

	case ROUTER_IFNEW:
		DEBUG_CFG("%s: ROUTER_IFNEW\n", cfg->devname);
		err = wan_device_new_if(wandev, cfg->arg);
		break;

	case ROUTER_IFDEL:
		DEBUG_CFG("%s: ROUTER_IFDEL\n", cfg->devname);
		err = wan_device_del_if(wandev, cfg->devname);
		break;

	case ROUTER_IFNEW_LIP:
		err = wan_device_new_if_lip(wandev,cfg->arg);
		break;

	case ROUTER_IFDEL_LIP:
		err = wan_device_del_if_lip(wandev,cfg->arg);
		break;

	case ROUTER_VER:
		DEBUG_CFG("%s: ROUTER_VER\n", cfg->devname);
		err = copyout((void*)wan_version,(void*)cfg->arg,strlen(wan_version)); 
		break;

	case ROUTER_IFSTAT:
        	break;

	case ROUTER_PROCFS:
        	DEBUG_CFG("wanpipe?: ROUTER_PROCFS\n");
		err = wan_device_procfs(wandev, cfg->arg);
		break;

	case ROUTER_DEBUG_READ:
        	DEBUG_CFG("wanpipe?: ROUTER_DEBUG_READ\n");
		if (wandev->debug_read){
			err = wandev->debug_read(wandev, cfg->arg);
		}else{
			err = -EINVAL;
		}
		break;

	case WANPIPE_DUMP:
	case WANPIPE_EXEC:
		DEBUG_CFG("%s: default\n", cfg->devname);
		if (wandev->ioctl){
			err = wandev->ioctl(wandev, cmd, cfg->arg);
		}else{
			DEBUG_EVENT("Unknow IOCTL call!!!\n");
			err = -EINVAL;
		}
		break;
	default:
		err = -EINVAL;
		break;
	}
	return err;
}

/*
 *	WAN Driver IOCTL Handlers
 */
/* 
 * ============================================================================
 *	Setup WAN link device.
 *	o verify user address space
 *	o allocate kernel memory and copy configuration data to kernel space
 *	o if configuration data includes extension, copy it to kernel space too
 *	o call driver's setup() entry point
 */

static int wan_device_setup (wan_device_t *wandev, wandev_conf_t *u_conf)
{
	void *data = NULL;
	wandev_conf_t *conf = NULL;
	int err = 0;

	DEBUG_CFG("%s: Device Setup...\n", wandev->name);
	if (wandev->setup == NULL){	/* Nothing to do ? */
		return 0;
	}

	conf = wan_malloc(sizeof(wandev_conf_t));
	if (conf == NULL){
		DEBUG_EVENT("%s: Cannot allocate memory for configuration structure!\n",
				wandev->name);
		return -ENOMEM;
	}

	if (copyin((void *)u_conf, (void*)conf, sizeof(wandev_conf_t))){
		wan_free(conf);
		return -EFAULT;
	}
	
	if (conf->magic != ROUTER_MAGIC){ 
        	DEBUG_EVENT("%s: Wrong configuration structure on Setup!\n", 
								wandev->name);	
		wan_free(conf);
		return -EINVAL; 
	}

	if (conf->data_size && conf->data){ 
		if (conf->data_size > 128000 || conf->data_size < 0){ 
			DEBUG_EVENT("%s: Large configuration structure\n", 
								wandev->name);
			wan_free(conf);
			return -EINVAL;
		}
		data = wan_malloc(conf->data_size);
		if (data){ 
			if (!copyin((void*)conf->data, data, conf->data_size)){
				conf->data = data;
				err = wandev->setup(wandev, conf);
			}else{
				err = -EFAULT;
			}
		}else{
			DEBUG_EVENT("%s: Cannot allocate data buffer\n", 
								wandev->name);
			err = -ENOMEM;
		}
		if (data){
			wan_free(data);
		}
	}else{
		if (conf->card_type == WANOPT_ADSL ||
		    conf->card_type == WANOPT_AFT ||
		    conf->card_type == WANOPT_AFT104 ||
		    conf->card_type == WANOPT_AFT300 ||
		    conf->config_id == WANCONFIG_DEBUG){
			conf->data = NULL;
			conf->data_size = 0;
			err = wandev->setup(wandev, conf);
		}else{
			DEBUG_EVENT("%s: ERROR: No firmware found (size = %i)!\n",
					wandev->name, conf->data_size);
		}
	}

	if (conf){
		wan_free(conf);
	}
	DEBUG_CFG("%s: Device setup finished (%d)\n", wandev->name, err);
	return err;
}

/* 
 * ============================================================================
 *	Shutdown WAN device.
 *	o delete all not opened logical channels for this device
 *	o call driver's shutdown() entry point
 */
 
static int wan_device_shutdown (wan_device_t *wandev, wandev_conf_t* u_conf)
{
	struct wan_dev_le	*devle, *tmp;
	wandev_conf_t		*conf = NULL;
	int 			error = 0, s;

	DEBUG_CFG("%s: device_shutdown called\n", wandev->name);	
	if (wandev->state == WAN_UNCONFIGURED){
		return 0;
	}

	if (u_conf){
		conf = wan_malloc(sizeof(wandev_conf_t));
		if (conf == NULL){
			DEBUG_EVENT("%s: Cannot allocate memory for configuration structure\n",
					wandev->name);
			return -ENOMEM;
		}

		if (copyin((void *)u_conf, (void*)conf, sizeof(wandev_conf_t))){
			wan_free(conf);
			return -EFAULT;
		}
		
		if (conf->magic != ROUTER_MAGIC){ 
        		DEBUG_EVENT("%s: Wrong configuration structure on Shutdown!\n", 
								wandev->name);	
			wan_free(conf);
			return -EINVAL; 
		}
	}

	/* Delete all interfaces */
	devle = WAN_LIST_FIRST(&wandev->dev_head);
	while(devle && devle->dev){
#if defined(__NetBSD__) && (__NetBSD_Version__ >= 106000200)
		s = splvm();
#else
		s = splimp(); 	/***** critical section start *****/
#endif
		WAN_LIST_REMOVE(devle, dev_link);
		--wandev->ndev;
		splx(s); 		
		if ((error = delete_interface(wandev, devle->dev))){
			return error;
		}
		tmp = devle;
		devle = WAN_LIST_NEXT(devle, dev_link);
		wan_free(tmp);
	}
	
	if (wandev->ndev){
		return -EBUSY;	/* there are opened interfaces  */
	}
		
	if (wandev->shutdown){
		error = wandev->shutdown(wandev, conf);
	}

	if (u_conf){

		if (copyout((void*)conf, (void*)u_conf, sizeof(wandev_conf_t))){
			return -EFAULT;
		}
		if (conf){
			wan_free(conf);
		}
	}

	DEBUG_CFG("%s: device_shutdown end\n", wandev->name);	
	return error;
}

/* 
 * ============================================================================
 *	Debugging WAN device.
 */
 
static int wan_device_debugging (wan_device_t *wandev, char* u_ifname)
{
	int 			error = 0;

	DEBUG_CFG("%s: device_debugging called\n", wandev->name);	
	if (wandev->state == WAN_UNCONFIGURED){
		return 0;
	}

	if (wandev->debugging){
		error = wandev->debugging(wandev);
	}

	DEBUG_CFG("%s: device_debugging end\n", wandev->name);	
	return 0;
}

/* 
 * ============================================================================
 *	Get WAN device status & statistics.
 */

static int wan_device_stat (wan_device_t *wandev, wandev_stat_t *u_stat)
{
	wandev_stat_t stat;

	bzero((void*)&stat, sizeof(stat));

	/* Ask device driver to update device statistics */
	if ((wandev->state != WAN_UNCONFIGURED) && wandev->update){
		wandev->update(wandev);
	}

	/* Fill out structure */
	stat.ndev  = wandev->ndev;
	stat.state = wandev->state;

	if (copyout((void*)&stat, (void*)u_stat, sizeof(stat))){
		return -EFAULT;
	}

	return 0;
}

/* 
 * ============================================================================
 *	Create new WAN interface.
 *	o verify user address space
 *	o copy configuration data to kernel address space
 *	o allocate network interface data space
 *	o call driver's new_if() entry point
 *	o make sure there is no interface name conflict
 *	o register network interface
 */

static int wan_device_new_if (wan_device_t *wandev, wanif_conf_t *u_conf)
{
    	netdevice_t	*ifp = NULL;
	sdla_t		*card = (sdla_t*)wandev->private;
	wanif_conf_t	conf;
    	int		interface_exist = 0;
    	int		err;

    	if ((wandev->state == WAN_UNCONFIGURED) || (wandev->new_if == NULL)){
		return -ENODEV;
	}

	if (copyin((void*)u_conf, (void*)&conf, sizeof(wanif_conf_t))){
		return -EFAULT;
	}
		
	if (conf.magic != ROUTER_MAGIC){
		return -EINVAL;
	}
    	DEBUG_EVENT("%s: Creating new WAN interface %s...\n", 
				wandev->name, conf.name);

	/* The network interface contanins of the interface in two forms:
	 * character string plus a unit number for the driver and binary
     	 * system wide index number. The user define the network interface 
	 * name, he responsible that these names must be different. The
	 * unit number will always 0 (I cannot do nothing with this.
	 */
	ifp = ifunit(conf.name);
	if (ifp){
		interface_exist = 1;    
	}else{
		if (wandev->config_id == WANCONFIG_ADSL){
			if(card->u.adsl.EncapMode == RFC_MODE_PPP_VC ||
			   card->u.adsl.EncapMode == RFC_MODE_PPP_LLC){
#if 1
				ifp = wan_netif_alloc(conf.name, WAN_IFT_PPP, &err);
#else
				ifp = (struct ifnet*)wan_malloc(sizeof(struct sppp));
#endif
			}else{
#if 1
				ifp = wan_netif_alloc(conf.name, WAN_IFT_ETHER, &err);
#else
				ifp = (struct ifnet*)wan_malloc(sizeof(wan_ethercom_t));
#endif
			}
		}else if (wandev->config_id == WANCONFIG_MPPP){
        		ifp = (struct ifnet*)wan_malloc(sizeof(struct sppp));
		}else{
#if 1
			ifp = wan_netif_alloc(conf.name, WAN_IFT_OTHER, &err);
#else
        		ifp = IFALLOC(WAN_IFT_OTHER);
#endif
#if 0
HDLC_SPPP
        		ifp = wan_malloc(sizeof(struct ifnet));
#endif
		}
		if (ifp == NULL){
			DEBUG_EVENT("%s: Failed allocate memory for interface %s!\n", 
						wandev->name, conf.name);
			return -ENOMEM;
		}
#if 0
		if (wandev->config_id == WANCONFIG_ADSL){
			if(card->u.adsl.EncapMode == RFC_MODE_PPP_VC ||
			   card->u.adsl.EncapMode == RFC_MODE_PPP_LLC){
				bzero((struct sppp*)ifp, sizeof(struct sppp));
			}else{
				bzero(WAN_IFP2AC(ifp), sizeof(wan_ethercom_t));
			}
		}else if (wandev->config_id == WANCONFIG_MPPP){
			bzero((struct sppp*)ifp, sizeof(struct sppp));
		}
		wan_netif_init((struct ifnet*)ifp, conf.name);
#endif
		interface_exist = 0;
	}
 
	err = wandev->new_if(wandev, ifp, &conf);
	if (!err){ 
		struct wan_dev_le	*devle;
		int			s;
		devle = wan_malloc(sizeof(struct wan_dev_le));
		if (devle == NULL){
			wandev->del_if(wandev, ifp);
			err = -ENOMEM;
			goto wan_device_new_if_error;
		}
		devle->dev = ifp;
#if defined(__NetBSD__) && (__NetBSD_Version__ >= 106000200)
		s = splvm();
#else
		s = splimp();
#endif
		WAN_LIST_INSERT_HEAD(&wandev->dev_head, devle, dev_link);
		++wandev->ndev;
		splx(s);
		if (!interface_exist){
			wanpipe_attach(ifp, wandev);
	    	}
		return 0;	/* done !!! */
	}
wan_device_new_if_error:
	wan_free(ifp);
	DEBUG_EVENT("%s: Failed to create new interface (%d)!\n",
						wandev->name, err);
	return err;
}


/* 
 * ============================================================================
 *	Delete WAN logical channel.
 *	 o verify user address space
 *	 o copy configuration data to kernel address space
 */
static int wan_device_del_if (wan_device_t *wandev, char *u_name)
{
	struct wan_dev_le	*devle;
	char 			name[WAN_IFNAME_SZ + 1];
	int			s, error;

	if (wandev->state == WAN_UNCONFIGURED){
		return -ENODEV;
	}
	
	bzero((void*)name, sizeof(name));

	if (copyin((void*)u_name, (void*)name, WAN_IFNAME_SZ)){
		return -EFAULT;
	}

	WAN_LIST_FOREACH(devle, &wandev->dev_head, dev_link){
		if (!strcmp(name, wan_netif_name(devle->dev))){
			break;
		}
	}
	if (devle == NULL || devle->dev == NULL){
		netdevice_t	*ifp;
		ifp = ifunit(name);
		if (ifp){
			DEBUG_EVENT("%s: Unregistering LIP interface\n", 
					name);
			return wan_device_del_if_lip(wandev,ifp);
		}else{
			DEBUG_EVENT("%s: Interface %s was not found (%s:%d)!\n",
					wandev->name, name,
					__FUNCTION__,__LINE__);
		}
		return -ENODEV;
	}
#if defined(__NetBSD__) && (__NetBSD_Version__ >= 106000200)
	s = splvm();
#else
	s = splimp(); 	/***** critical section start *****/
#endif
	WAN_LIST_REMOVE(devle, dev_link);
	--wandev->ndev;
	splx(s); 		
	if ((error = delete_interface(wandev, devle->dev))){
		DEBUG_EVENT("%s: Failed to delete interface %s (%s:%d)!\n",
				wandev->name, u_name,
				__FUNCTION__,__LINE__);
		return error;
	}
	wan_free(devle);
	return error;
}


/*
 *	Miscellaneous Functions
 */

/* 
 * ===========================================================================
 *	Find WAN device by name.
 *	Return pointer to the WAN device data space or NULL if device not found.
 */

static wan_device_t *find_device(char *name)
{
	wan_device_t *wandev = NULL;

	WAN_LIST_FOREACH(wandev, &wan_devlist, next){
		if (!strcmp(wandev->name, name)){
			break;
		}
	}
/*
	for (wandev = router_devlist;wandev && strcmp(wandev->name, name);
		wandev = wandev->next);
*/
	return wandev;
}

/* 
 * ============================================================================
 *	Delete WAN logical channel identified by its name.
 *	o find logical channel by its name
 *	o call driver's del_if() entry point
 *	o unregister network interface
 *	o unlink channel data space from linked list of channels
 *	o release channel data space
 *
 *	Return:	0		success
 *		-ENODEV		channel not found.
 *		-EBUSY		interface is open
 *
 *	Note: If (force != 0), then device will be destroyed even if interface
 *	associated with it is open. It's caller's responsibility to make
 *	sure that opened interfaces are not removed!
 */

static int delete_interface (wan_device_t *wandev, netdevice_t *dev)
{
	int err;

	DEBUG_CFG("%s: Deleting interface %s...\n", 
			wandev->name, wan_netif_name(dev));
	if ((dev->if_flags & IFF_UP) == IFF_UP){ 
		dev->if_flags &= ~IFF_UP;
	}

	/* Unregister the lip interface attached
	** to this interace */
	err=wan_device_unreg_lip(dev);
	if (err){
		return err;
	}

	if (wandev->del_if){
		wandev->del_if(wandev, dev);
	}
	wanpipe_detach(dev, wandev);
    
#if (defined __FreeBSD__) && (__FreeBSD_version < 400000)
/*
 * Do not enable IF_DETACH for FreeBSD-3.x (kernel bug). 
 */
# ifdef IF_DETACH
	wan_free(dev);
# endif
#else
	wan_free(dev);
#endif
	return 0;
}

static void wanpipe_attach(struct ifnet *ifp, wan_device_t* wandev)
{
    	WAN_IFQ_SET_READY(&ifp->if_snd);
	if (ifunit(wan_netif_name(ifp))){
		DEBUG_CFG("%s:%s: Network interface is already attached to kernel!\n",
				wandev->name, wan_netif_name(ifp));
		return;
	}
	
	DEBUG_EVENT("%s: Attaching network interface %s...\n", 
					wandev->name, wan_netif_name(ifp));
	if (wandev->config_id == WANCONFIG_ADSL){
		sdla_t*	card = (sdla_t*)wandev->private;
		switch(card->u.adsl.EncapMode){
		case RFC_MODE_PPP_VC:
		case RFC_MODE_PPP_LLC:
    			IFATTACH(ifp);
    			BPF_ATTACH(ifp, DLT_PPP);
			break;
		case RFC_MODE_BRIDGED_ETH_LLC:	
		case RFC_MODE_BRIDGED_ETH_VC:
			ETHER_IFATTACH(ifp, wandev->macAddr);
			break;
		case RFC_MODE_ROUTED_IP_LLC:
		case RFC_MODE_ROUTED_IP_VC:	
    			IFATTACH(ifp);
    			BPF_ATTACH(ifp, DLT_RAW);
			break;
		}
	}else{
    		IFATTACH(ifp);
    		BPF_ATTACH(ifp, (wandev->config_id == WANCONFIG_MPPP) ? DLT_PPP : DLT_RAW);
	}
}

static void wanpipe_detach(struct ifnet *ifp, wan_device_t* wandev)
{
	if (!ifunit(wan_netif_name(ifp))){
		return;
	}
#if defined(__FreeBSD__)
#if  (__FreeBSD_version < 400000)
    	struct ifaddr* ifa = NULL, *dst_ifa = NULL;
    	int s = splnet();

    	/* Remove addresses from routing tables */
    	ifa = TAILQ_FIRST(&ifp->if_addrhead);
    	for(; ifa; ifa = TAILQ_NEXT(ifa, ifa_link)){
		rtinit(ifa, (int)RTM_DELETE, 0);
	}
    	if ((ifa = TAILQ_FIRST(&ifp->if_addrhead)) != NULL){
		if ((dst_ifa = TAILQ_NEXT(ifa, ifa_link)) != NULL){
	    		TAILQ_REMOVE(&ifp->if_addrhead, dst_ifa, ifa_link);
		}
	}
/*
 * Do not enable IF_DETACH for FreeBSD-3.x (kernel bug). 
 */
#  ifdef IF_DETACH
    	TAILQ_REMOVE(&ifnet, ifp, if_link);
#  endif 
    	splx(s);
# endif /* __FreeBSD_version */
#endif  /* __FreeBSD__ */

	if (wandev->config_id == WANCONFIG_ADSL){
		sdla_t*	card = (sdla_t*)wandev->private;
		switch(card->u.adsl.EncapMode){
		case RFC_MODE_PPP_VC:
		case RFC_MODE_PPP_LLC:
    			IFDETACH(ifp);
			break;
		case RFC_MODE_BRIDGED_ETH_LLC:	
		case RFC_MODE_BRIDGED_ETH_VC:
			ETHER_IFDETACH(ifp);
			break;
		case RFC_MODE_ROUTED_IP_LLC:
		case RFC_MODE_ROUTED_IP_VC:	
    			IFDETACH(ifp);
			break;
		}
		wan_netif_del(ifp);
		if (ifp->if_softc){
			wan_free(ifp->if_softc);
			ifp->if_softc = NULL;
		}
	}else{
    		IFDETACH(ifp);
	}
}

static int wan_device_procfs(wan_device_t* wandev, wan_procfs_t* u_procfs)
{
	wan_procfs_t	procfs;
	char*		data = NULL;
	int		cnt = 0;

	if (copyin((void*)u_procfs, (void*)&procfs, sizeof(wan_procfs_t))){
		return -EFAULT;
	}
		
	if (procfs.magic != ROUTER_MAGIC){
		return -EINVAL;
	}
	if ((data = wan_malloc(procfs.max_len)) == NULL){
		return -ENOMEM;
	}
	switch(procfs.cmd){
	case WANPIPE_PROCFS_CONFIG:
		DEBUG_PROCFS("WANPIPE_PROCFS_CONFIG!\n");
		cnt = config_get_info(data, NULL, procfs.offs, procfs.max_len);
		break;	

	case WANPIPE_PROCFS_HWPROBE:
		DEBUG_PROCFS("WANPIPE_PROCFS_HWPROBE!\n");
		cnt = probe_get_info(data, NULL, procfs.offs, procfs.max_len);
		break;	

	case WANPIPE_PROCFS_STATUS:
		DEBUG_PROCFS("WANPIPE_PROCFS_STATUS!\n");
		cnt = status_get_info(data, NULL, procfs.offs, procfs.max_len);
		break;	

	case WANPIPE_PROCFS_INTERFACES:
		DEBUG_PROCFS("WANPIPE_PROCFS_INTERFACES!\n");
		cnt = interfaces_get_info(data, NULL, procfs.offs, procfs.max_len);
		break;	

	case WANPIPE_PROCFS_DEV:
#if 0
		DEBUG_PROCFS("WANPIPE_PROCFS_DEV!\n");
		if (wandev){
			cnt = wandev_get_info(data, wandev, procfs.offs, procfs.max_len);
		}
#endif
		break;	

	}
	procfs.is_more = (cnt) ? 1 : 0;
	procfs.offs = cnt;
	if (cnt){
		if (copyout((void*)data, (void*)procfs.data, cnt)){
			return -EFAULT;
		}
		if (data){
			wan_free(data);
		}
	}
	if (copyout((void*)&procfs, (void*)u_procfs, sizeof(wan_procfs_t))){
		return -EFAULT;
	}
	return 0;
}

int register_wanpipe_lip_protocol (wplip_reg_t *lip_reg)
{
	DEBUG_EVENT("Registering Wanpipe Lip protocol\n");
	memcpy(&wplip_protocol,lip_reg,sizeof(wplip_reg_t));
	REG_PROTOCOL_FUNC(wplip_protocol);
	return 0;
}

void unregister_wanpipe_lip_protocol (void)
{
	UNREG_PROTOCOL_FUNC(wplip_protocol);
	return;
}

static int wan_device_unreg_lip(netdevice_t *dev)
{
	void *lip_link = wan_get_lip_ptr(dev);
	int err;

	if (!IS_FUNC_CALL(wplip_protocol,wplip_if_unreg))
		return 0;

	if (lip_link){
		err=wplip_protocol.wplip_unreg(lip_link);
		if (err){
			return err;
		}
		wan_set_lip_ptr(dev,NULL);
	}
	return 0;
}

static int wan_device_new_if_lip (wan_device_t *wandev, wanif_conf_t *u_conf)
{
	wanif_conf_t *conf=NULL;
	netdevice_t *dev=NULL;
	netdevice_t *tmp_dev;
	void *lip_link;
	int err, init_if=0;

	if ((wandev->state == WAN_UNCONFIGURED) || (wandev->new_if == NULL))
		return -ENODEV;
	
	conf=wan_malloc(sizeof(wanif_conf_t));
	if (!conf){
		return -ENOMEM;
	}
	
	if(WAN_COPY_FROM_USER(conf, u_conf, sizeof(wanif_conf_t))){
		DEBUG_EVENT("%s: Failed to copy from user space!\n",
				wandev->name);
		err = -EFAULT;
		goto wan_device_new_if_lip_exit;
	}
		
	if (conf->magic != ROUTER_MAGIC){
		DEBUG_EVENT("%s: Wrong magic number!\n", 
				wandev->name);
		err = -EINVAL;
		goto wan_device_new_if_lip_exit;
	}

	if (!wan_test_bit(0,&wplip_protocol.init)){
		DEBUG_EVENT("%s: LIP protocol not initialized!\n",
				wandev->name);
		err = -EPROTONOSUPPORT;
		goto wan_device_new_if_lip_exit;
	}


	err = -EPROTONOSUPPORT;

	if (!conf->name){
		DEBUG_EVENT("wanpipe: NEW_IF LIP no interface name!\n");
		err = -EINVAL;	
		goto wan_device_new_if_lip_exit;
	}

	if ((tmp_dev=ifunit(conf->name)) != NULL){
		DEBUG_EVENT("%s: Device already exists!\n",
				conf->name);
		err = -EEXIST;
		goto wan_device_new_if_lip_exit;
	}

	DEBUG_EVENT("%s: Registering LIP %s -> %s\n",
			wandev->name, conf->name, conf->master);

	if (!conf->master){
		DEBUG_EVENT("wanpipe: NEW_IF LIP no master interface name!\n");
		err = -EINVAL;	
		goto wan_device_new_if_lip_exit;
	}
		
	/* Find a master device lip device */
	if ((dev = ifunit(conf->master)) == NULL){
		DEBUG_EVENT("%s: Master device %s used by LIP %s no found!\n",
				wandev->name,conf->master,conf->name);
		goto wan_device_new_if_lip_exit;
	}

	lip_link = (wplip_reg_t*)wan_get_lip_ptr(dev); 
	if (!lip_link){

		err=wplip_protocol.wplip_register(&lip_link,conf, wan_netif_name(dev));
		if (err!=0){
			DEBUG_EVENT("%s: Failed to register LIP Link device %s\n",
					wandev->name,conf->name);
			goto wan_device_new_if_lip_exit;
		}
	
	
		err=wplip_protocol.wplip_bind_link(lip_link,dev);
		if (err!=0){
			wplip_protocol.wplip_unreg(lip_link);
			wan_set_lip_ptr(dev,NULL);

			DEBUG_EVENT("%s: Failed to bind master dev(%s) to LIP Link device\n",
					wandev->name, wan_netif_name(dev));
			goto wan_device_new_if_lip_exit;
		}
		
		wan_set_lip_ptr(dev,lip_link);
		init_if=1;
	}

	err=wplip_protocol.wplip_if_reg(lip_link,conf->name,conf);
	if (err != 0){	
		DEBUG_EVENT("%s: Failed to register LIP device %s\n",
				wandev->name,conf->name);

		if (init_if){
			wplip_protocol.wplip_unreg(lip_link);
			wan_set_lip_ptr(dev,NULL);
		}
		
		goto wan_device_new_if_lip_exit;
	}

wan_device_new_if_lip_exit:

	if (conf){
		wan_free(conf);
	}
	
	return err;
}

static int wan_device_del_if_lip(wan_device_t *wandev, netdevice_t *dev)
{
	int err;
	
	if (dev->if_flags & IFF_UP){
		DEBUG_EVENT("%s: Failed to del interface: Device UP!\n",
				wan_netif_name(dev));
		return -EBUSY;
	}
	
	if (!IS_FUNC_CALL(wplip_protocol,wplip_if_unreg))
		return 0;

	err=wan_device_unreg_lip(dev);
	if (err){
		return err;
	}
	
	return wplip_protocol.wplip_if_unreg(dev);
}

/*================================================
 * Interface to LIP Layer
 */
int wanpipe_lip_rx(void *chan_ptr, void *skb)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	
	if (!chan->lip){
		return -ENODEV;
	}
	if (!IS_FUNC_CALL(wplip_protocol,wplip_rx))
		return -EPROTONOSUPPORT;

	return wplip_protocol.wplip_rx(chan->lip,skb);
}

int wanpipe_lip_connect(void *chan_ptr, int reason)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	
	if (!chan->lip){
		return -ENODEV;
	}
	if (!IS_FUNC_CALL(wplip_protocol,wplip_connect))
		return -EPROTONOSUPPORT;

	 wplip_protocol.wplip_connect(chan->lip,reason);

	 return 0;
}

int wanpipe_lip_disconnect(void *chan_ptr, int reason)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	
	if (!chan->lip){
		return -ENODEV;
	}
	if (!IS_FUNC_CALL(wplip_protocol,wplip_disconnect))
		return -EPROTONOSUPPORT;

	wplip_protocol.wplip_disconnect(chan->lip,reason);
	return 0;
}

int wanpipe_lip_kick(void *chan_ptr,int reason)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	
	if (!chan->lip){
		return -ENODEV;
	}
	if (!IS_FUNC_CALL(wplip_protocol,wplip_kick))
		return -EPROTONOSUPPORT;

	wplip_protocol.wplip_kick(chan->lip,reason);
	return 0;
}

int wanpipe_lip_get_if_status(void *chan_ptr,void *m)
{
	wanpipe_common_t *chan = (wanpipe_common_t*)chan_ptr;
	
	if (!chan->lip){
		return -ENODEV;
	}
	if (!IS_FUNC_CALL(wplip_protocol,wplip_get_if_status))
		return -EPROTONOSUPPORT;

	wplip_protocol.wplip_get_if_status(chan->lip,m);
	return 0;


}


/************************ End of Wanmain.c ******************************/
